/*
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.tbgf.games.desdemona.client;

import java.util.Vector;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import net.sf.goodgame.shared.InvalidConfigurationException;
import net.sf.goodgame.shared.games.GameOptions;
import net.sf.goodgame.shared.games.Player;
import net.sf.goodgame.tbgf.client.TurnBasedGame;
import net.sf.goodgame.tbgf.client.TurnBasedGameClient;
import net.sf.goodgame.tbgf.games.desdemona.shared.Move;
import net.sf.goodgame.tbgf.games.desdemona.shared.Piece;

/**
 *
 * @author mchaten
 */
public class DesdemonaClientEngine implements TurnBasedGame {

    private DesdemonaUI ui;
    private byte boardSize;
    private DesdemonaClient gameClient;
    private Vector<Move> allowableMoves;
    private Player currentTurn;
    private Player blackPlayer;
    private Player whitePlayer;
    private Player me;
    private boolean gameStarted;

    public DesdemonaClientEngine() {
        gameClient = null;
        boardSize = -1;
        ui = null;
        allowableMoves = new Vector<Move>();
    }

    public void gameOver(String winner) {
        ui.gameOver(winner);
    }

    @Override
    public void gameStarted(Player me, GameOptions options) {
        String white = options.getAsString("white");
        String black = options.getAsString("black");
        this.me = me;
        whitePlayer = new Player(white);
        blackPlayer = new Player(black);
        currentTurn = whitePlayer;
        gameStarted = true;
        if(currentTurn.equals(me))
        {
            generateAllowableMoves();
        }
    }

    @Override
    public JPanel getGamePanel() {
        return ui;
    }

    @Override
    public void moveTimeout() {
    }

    @Override
    public void playerJoined(Player player) {
    }

    @Override
    public void playerLeft(String player) {
    }

    @Override
    public void gameMessage(String command) {
        System.out.println("PrintMove(String) was called, This Should Never happen.");
    }

    @Override
    public void gameMessage(byte[] command, short read) {
        System.out.println("Move Read");
        Move move = Move.getMove(command, read);
        if (move == null) {
            System.out.println("Move Invalid");
            return;
        }
        updateBoard(move);
        return;
    }

    private void updateBoard(Move move) {
        //We assume the move is a valid because the server checks it.
        ui.setPiece(move.X, move.Y, currentTurn());
        for (int index = 0; index < move.getIndex(); index++) {
            updateDirection(move.X, move.Y, move.getDeltaX(index), move.getDeltaY(index));
        }
    }

    private void updateDirection(byte x, byte y, byte deltaX, byte deltaY) {
        for (x += deltaX   , y += deltaY;true; x += deltaX, y += deltaY) {
            if (ui.getPiece(x, y) == currentTurn()) {
                return;
            } else {
                ui.setPiece(x, y, currentTurn());
            }
        }

    }

    private Piece currentTurn() {
        return isBlacksTurn() ? Piece.black : Piece.white;
    }

    private Piece otherTurn() {
        return isBlacksTurn() ? Piece.white : Piece.black;
    }

    private boolean isCurrentTurnsPiece(Piece piece) {
        return currentTurn() == piece;
    }

    private boolean isBlacksTurn() {
        if (gameStarted) {
            return blackPlayer.equals(currentTurn);
        }
        return false;
    }

    @Override
    public void setGameClient(TurnBasedGameClient client) {
        gameClient = (DesdemonaClient) client;
    }

    @Override
    public void initializeUI() {
        ui = new DesdemonaUI(this, gameClient);
    }

    private void generateAllowableMoves() {
        allowableMoves.clear();
        for (byte x = 0; x < ui.getBoardSize(); x++) {
            for (byte y = 0; y < ui.getBoardSize(); y++) {
                Piece piece = ui.getPiece(x, y);
                //Check each of the 8 directions
                if (isCurrentTurnsPiece(piece)) {
                    for (byte deltaX = -1; deltaX <= 1; deltaX++) {
                        for (byte deltaY = -1; deltaY <= 1; deltaY++) {
                            if (deltaX == 0 && deltaY == 0) {
                                continue;
                            }
                            Move move = getMovesInDirection(x, y, deltaX, deltaY);
                            if (move != null) {
                                addMove(move);
                            }
                        }
                    }
                }
            }
        }
        ui.showValidMoves(allowableMoves.toArray(new Move[0]));
    }

    private void addMove(Move move) {
        for (Move m : allowableMoves) {
            if (m != null && m.X == move.X && m.Y == move.Y) {
                m.setDirection(move.getDeltaX(0), move.getDeltaY(0));
                return;
            }
        }
        allowableMoves.add(move);
    }
    /*
    private Move getMovesInDirection(byte x, byte y, byte deltaX, byte deltaY) {
    for (y += deltaY,x+=deltaX;x>=0 && x < ui.getBoardSize() && y >= 0 && y < ui.getBoardSize(); x += deltaX, y += deltaY) {
    Piece piece = ui.getPiece(x, y);
    boolean yBounds = (y - deltaY >= 0 && y - deltaY < ui.getBoardSize());
    boolean xBounds = (x - deltaX >= 0 && x - deltaX < ui.getBoardSize());
    boolean bounds = yBounds && xBounds;
    if (isCurrentTurnsPiece(piece) || (bounds && ui.getPiece((byte) (x - deltaX), (byte) (y - deltaY)) == null)) {
    return null;
    } else if (piece == null && bounds && !isCurrentTurnsPiece(ui.getPiece((byte) (x - deltaX), (byte) (y - deltaY)))) {
    Move move = new Move(x, y);
    move.setDirection((byte) -deltaX, (byte) -deltaY);
    return move;
    }
    }
    return null;
    }*/

    private Move getMovesInDirection(byte x, byte y, byte deltaX, byte deltaY) {
        Piece piece = ui.getPiece(x, y);
        x += deltaX;
        y += deltaY;
        if (piece == currentTurn() && x >= 0 && y >= 0 && x < ui.getBoardSize() && y < ui.getBoardSize()) {
            piece = ui.getPiece(x, y);
            if (piece == otherTurn()) {
                boolean bounds = true;
                do {
                    piece = ui.getPiece(x, y);
                    if (piece != otherTurn()) {
                        break;
                    }
                    x += deltaX;
                    y += deltaY;
                    bounds = x >= 0 && x < ui.getBoardSize() && y >= 0 && y < ui.getBoardSize();
                } while (bounds);
                if (piece == null && bounds) {
                    Move m = new Move(x, y);
                    m.setDirection((byte) (-deltaX), (byte) (-deltaY));
                    return m;
                }
            }
        }
        return null;
    }
    // */

    public void setSize(byte size) {

        this.boardSize = size;
        try {
            ui.setBoardSize(size);
        } catch (InvalidConfigurationException e) {
            throw new RuntimeException(e.getMessage());
        }

    }

    public void tryMove(byte x, byte y) {
        if (!gameStarted) {
            JOptionPane.showMessageDialog(ui, "The game has not started yet!");
            return;
        }

        if (!currentTurn.equals(me)) {
            //TODO: Change to something not so anoying.
            JOptionPane.showMessageDialog(ui, "Not your turn!");
            return;

        }
        if (allowableMoves.size() == 0) {
            generateAllowableMoves();
        }

        for (Move m : allowableMoves) {
            if (m.X == x && m.Y == y) {
                //Send do that move and return
                this.gameClient.sendMessage(m.getNetworkMessage());
                return;
            }
        }
    //Reach here if we didn't find a valid move.
    }

    public void turnOver(Player nextTurn) {
        System.out.println("Next turn: " + nextTurn);
        ui.clearValidMoves();
        currentTurn = nextTurn;
        if (currentTurn.equals(me)) {
            generateAllowableMoves();
        }
    }

    public void spectate(Piece[][] gameBoard, Player white, Player black, Player currentTurn, Player me) {
        System.out.println("Spec called");
        whitePlayer = white;
        blackPlayer = black;
        this.me = me;
        this.currentTurn = currentTurn;
        for (byte i = 0; i < gameBoard.length; i++) {
            for (byte j = 0; j < gameBoard[i].length; j++) {
                if (gameBoard[i][j] == null) {
                    continue;
                }
                ui.setPiece(i, j, gameBoard[i][j]);
            }
        }
        gameStarted = true;
    }
}
