/*
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.tbgf.games.desdemona.server;

import net.sf.goodgame.tbgf.server.*;
import java.util.Vector;
import net.sf.goodgame.server.game.GameResult;
import net.sf.goodgame.shared.InvalidConfigurationException;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.games.GameOptions;
import net.sf.goodgame.shared.games.Player;
import net.sf.goodgame.tbgf.games.desdemona.shared.GameBoard;
import net.sf.goodgame.tbgf.games.desdemona.shared.Move;
import net.sf.goodgame.tbgf.games.desdemona.shared.Piece;

/**
 *
 * @author mchaten
 */
public class DesdemonaServerEngine implements TurnBasedGame {

    private Player currentTurn;
    private Vector<Move> allowableMoves;
    private volatile boolean gameStarted;
    private Player blackPlayer;
    private Player whitePlayer;
    private GameBoard gameBoard;
    private DesdemonaServer network;

    public DesdemonaServerEngine(DesdemonaServer server) {
        gameStarted = false;
        allowableMoves = new Vector<Move>();
        network = server;
    }

    public synchronized boolean isPlayersTurn(Player player) {
        return player.equals(currentTurn) && gameStarted;
    }

    public synchronized void moveTimeout() {
        if (gameStarted) {
            playerMove(allowableMoves.firstElement());
        }
    }

    public synchronized void playerJoined(Player player) {
        
        if (gameStarted) {
            //Spectator Initialize
            byte[][] gb = new byte[gameBoard.getSize()][gameBoard.getSize()];
            for(byte i = 0;i<gb.length;i++)
            {
                for(byte j = 0;j<gb[i].length;j++)
                {
                    Piece p = gameBoard.getPiece(i, j);
                    gb[i][j] = p==null?-1:(byte)p.ordinal();
                }
            }
            GameOptions spec = new GameOptions();
            spec.put("spectator",Util.combineArrays(gb));
            spec.put("black",blackPlayer.getUsername());
            spec.put("white",whitePlayer.getUsername());
            spec.put("current",currentTurn.getUsername());
            network.setGameOptions(spec, player);
            network.sendGameOptions(player);
        } else {
            //Player Initialize
            network.sendGameIsSetup(player);
            GameOptions options = new GameOptions();
            if (whitePlayer == null) {
                whitePlayer = player;
                options.put("white",player.getUsername());
                network.setGameOptions(options);
            } else {
                blackPlayer = player;
                options.put("black",player.getUsername());
                network.setGameOptions(options);
                gameStarted();
            }
        }
    }

    private void gameStarted() {
        GameOptions options = network.getGameOptions();
        byte size = options.getAsByte("size");
        try {
            gameBoard = new GameBoard(size);
            generateAllowableMoves();
        } catch (InvalidConfigurationException e) {
            throw new RuntimeException(e.getMessage());
        }
        currentTurn = whitePlayer;
        gameStarted = true;
        network.gameStarted();
        
    }

    public synchronized void playerLeft(Player player) {
        if (gameStarted) {
            GameResult r = new GameResult(new Player[]{},new Player[]{player});
            //TODO: Check if the game has reached a certain point, and if it has, let the guy who didn't leave win.
            gameOver(r,player+" left, he loses");
        } else {
            if(player.equals(whitePlayer))
            {
                whitePlayer = null;
            }
            else if(player.equals(blackPlayer))
            {
                blackPlayer = null;
            }
        }
    }

    public synchronized void gameMessage(Player p, String command) {
        return; //Not called
    }

    private void playerMove(Move curMove) {
        //It is assumed that the move refers to the currentTurn
        //Check if the move is valid
        for (Move m : allowableMoves) {
            if (m.X == curMove.X && m.Y == curMove.Y) {
                doMove(m);
                network.sendMessage(m.getNetworkMessage());
                turnOver();
                return;
            }
        }
    }

    private void doMove(Move move) {
        gameBoard.setPiece(move.X, move.Y, currentTurn());
        for (int index = 0; index < move.getIndex(); index++) {
            updateDirection(move.X, move.Y, move.getDeltaX(index), move.getDeltaY(index));
        }
        return;
    }

    private void updateDirection(byte x, byte y, byte deltaX, byte deltaY) {
        for (x += deltaX   , y += deltaY;true; x += deltaX, y += deltaY) {
            if (gameBoard.getPiece(x, y) == currentTurn()) {
                return;
            } else {
                gameBoard.setPiece(x, y, currentTurn());
            }
        }

    }

    public synchronized Player getCurrentTurn() {
        return currentTurn;
    }

    public synchronized void gameMessage(Player player, byte[] command, short read) {
        if (gameStarted && isPlayersTurn(player)) {
            Move move = Move.getMove(command, read);
            if (move == null) {
                return;
            }
            playerMove(move);
            return;
        }
    }

    private void turnOver() {
        currentTurn = (currentTurn == whitePlayer) ? blackPlayer : whitePlayer;
        //TODO: Perhaps replace this with an algorithm that checks if a move is valid rather than compiling a list of valid moves?
        generateAllowableMoves();
        if (allowableMoves.size() == 0) {
            gameOver();
        } else {
            network.turnOver(currentTurn);
        }
    }

    private void generateAllowableMoves() {
        allowableMoves.clear();
        for (byte x = 0; x < gameBoard.getSize(); x++) {
            for (byte y = 0; y < gameBoard.getSize(); y++) {
                Piece piece = gameBoard.getPiece(x, y);
//                Check each of the 8 directions
                if (isCurrentTurnsPiece(piece)) {
                    for (byte deltaX = -1; deltaX <= 1; deltaX++) {
                        for (byte deltaY = -1; deltaY <= 1; deltaY++) {
                            if (deltaX == 0 && deltaY == 0) {
                                continue;
                            }
                            Move move = getMovesInDirection(x, y, deltaX, deltaY);
                            if (move != null) {
                                addMove(move);
                            }
                        }
                    }
                }
            }
        }
    }

    private void addMove(Move move) {
        for (Move m : allowableMoves) {
            if (m != null && m.X == move.X && m.Y == move.Y) {
                m.setDirection(move.getDeltaX(0), move.getDeltaY(0));
                return;
            }
        }
        allowableMoves.add(move);
    }

    private Move getMovesInDirection(byte x, byte y, byte deltaX, byte deltaY) {
        y += deltaY;
        x += deltaX;
        for (; x >= 0 && x < gameBoard.getSize() && y >= 0 && y < gameBoard.getSize(); x += deltaX, y += deltaY) {
            Piece piece = gameBoard.getPiece(x, y);
            boolean yBounds = (y - deltaY >= 0 && y - deltaY < gameBoard.getSize());
            boolean xBounds = (x - deltaX >= 0 && x - deltaX < gameBoard.getSize());
            boolean bounds = xBounds & yBounds;
            if (isCurrentTurnsPiece(piece) || (bounds && gameBoard.getPiece((byte) (x - deltaX), (byte) (y - deltaY)) == null)) {
                return null;
            } else if (piece == null && bounds && !isCurrentTurnsPiece(gameBoard.getPiece((byte) (x - deltaX), (byte) (y - deltaY)))) {
                Move move = new Move(x, y);
                move.setDirection((byte) -deltaX, (byte) -deltaY);
                return move;
            }
        }
        return null;
    }

    private boolean isCurrentTurnsPiece(Piece piece) {
        return (isBlacksTurn() && piece == Piece.black) || (!isBlacksTurn() && piece == piece.white);
    }

    private boolean isBlacksTurn() {
        return currentTurn == blackPlayer;
    }

    private Piece currentTurn() {
        return isBlacksTurn() ? Piece.black : Piece.white;
    }

    private void gameOver() {
        Piece winner = gameBoard.hasMostPieces();
        if (winner == null) {
            GameResult r = new GameResult(new Player[]{blackPlayer,whitePlayer});
            gameOver(r,"Tie!");
        } else if (winner == Piece.black) {
            GameResult r = new GameResult(new Player[]{blackPlayer},new Player[]{whitePlayer});
            gameOver(r,"Black("+blackPlayer.getUsername()+") wins!");
        } else {
            GameResult r = new GameResult(new Player[]{whitePlayer},new Player[]{blackPlayer});
            gameOver(r,"White("+whitePlayer.getUsername()+") wins!");
        }
    }
    private void gameOver(GameResult r,String message)
    {
        gameStarted = false;
        network.gameOver(r,message);
    }
}
