/*    
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.tbgf.client;

import java.util.logging.Level;
import java.util.logging.Logger;
import net.sf.goodgame.tbgf.shared.ServerToClientProtocol;
import javax.swing.JPanel;
import net.sf.goodgame.client.game.GameClient;
import net.sf.goodgame.client.game.GameClientListener;
import net.sf.goodgame.client.gui.chat.ChatBox;
import net.sf.goodgame.client.gui.chat.ChatListener;
import net.sf.goodgame.client.gui.translucency.TranslucentJPanel;
import net.sf.goodgame.client.networking.InputOutput;
import net.sf.goodgame.client.user.User;
import net.sf.goodgame.client.user.UserList;
import net.sf.goodgame.client.user.UserListGUI;
import net.sf.goodgame.client.user.UserListListener;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.games.GameOptions;
import net.sf.goodgame.shared.games.Player;
import net.sf.goodgame.shared.network.InvalidMessageException;
import net.sf.goodgame.tbgf.shared.ClientToServerProtocol;

public abstract class TurnBasedGameClient implements GameClient, UserListListener {

    protected GameClientListener gameListener;
    protected ChatListener chatListener;
    protected TurnBasedGame game;
    protected JPanel rootPanel;
    protected ChatBox chat;
    protected UserList userList;
    protected InputOutput io;
    private Player me;

    //This should return the title of your game.
    protected abstract String getGameTitle();

    //Set game to be the class that this client is for.
    protected abstract void initializeGame();

    //Is called when a user trys to make a new game.
    protected abstract GameOptions setupNewGame(GameOptions options);

    protected abstract void gameIsSetup(GameOptions options);
    //Stop any game threads or anything that would make this stay around when
    //Not used.

    public abstract void destroy();

    //Typically only one of these needs to be implemented
    //If you call gameOver(String) on the server, gameOver(String)
    //Will be called here and likewise for byte[]
    public abstract void gameOver(String message);

    public abstract void gameOver(byte[] message);

    //Add the components to the rootPanel variable that are used for the game.
    protected abstract void initializeUI();

    //Typically only called once, default is to return the gameOptions from
    //the GameSetup object returned from setupNewGame(JPanel)
    protected abstract void setGameOptions(GameOptions options);

    public TurnBasedGameClient() {
        rootPanel = new TranslucentJPanel(.5f);

    }

    public void gameMessage(String message) {
        return; //Never going to be called
    }

    public void gameMessage(byte[] message) {
        ServerToClientProtocol messageType = ServerToClientProtocol.valueOf(message[0]);
        short read = 1;
        switch (messageType) {
            case setupGameOptions:
                setupGameOptions(message, read);
                break;
            case gameIsSetup:
                gameIsSetup(message, read);
                break;
            case moveTimeout:
                game.moveTimeout();
                break;
            case gameMessageBinary:
                playerMoveBinary(message, read);
                break;
            case gameMessageString:
                playerMoveString(message, read);
                break;
            case gameStarted:
                gameStarted(message);
                break;
            case gameOptions:
                try {
                    setGameOptions(GameOptions.getGameOptions(message, (short) 1));
                } catch (InvalidMessageException e) {
                }
                break;
            case turnOver:
                String playerName = new String(Util.extractArray(message, 1, message.length - 1));
                game.turnOver(new Player(playerName));
                break;
            case gameOverString:
                gameOver(new String(message));
                break;
            case gameOverBinary:
                gameOver(message);
                break;

            default:
                System.out.println("Invalid message recieved");
                break;
        }
    }

    private void gameIsSetup(byte[] message, short read) {
       try {
            GameOptions options = GameOptions.getGameOptions(message, read);
            gameIsSetup(options);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(TurnBasedGameClient.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private void setupGameOptions(byte[] message, short read) {
        try {
            GameOptions options = GameOptions.getGameOptions(message, read);
            options = setupNewGame(options);
            send(ClientToServerProtocol.setupGame,options.getNetworkMessage());
        } catch (InvalidMessageException ex) {
            Logger.getLogger(TurnBasedGameClient.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public JPanel getGamePanel() {
        return rootPanel;
    }

    public void setGameClientListener(GameClientListener lgl) {
        gameListener = lgl;
        me = gameListener.whoAmI();
        io = gameListener.getIO();
        io.getIncomingConnection().setGameClient(this);
        chat = new ChatBox(getGameTitle(), io.getOutgoingConnection(), false);
        io.getIncomingConnection().setChatListener(chat);
        userList = new UserListGUI(gameListener.getPrivateChatController(), io.getOutgoingConnection());
        userList.addListener(this);
        io.getIncomingConnection().setUserList(userList);
        initializeGame();
        initializeUI();
    }

    private void playerMoveString(byte[] message, short read) {
        byte[] move = Util.extractArray(message, read, message.length - read);
        game.gameMessage(new String(move));
    }

    private void playerMoveBinary(byte[] message, short read) {
        game.gameMessage(message, read);
    }

    public void sendMessage(byte[] message) {
        send(ClientToServerProtocol.gameMessageBinary, message);
    }

    public void sendMessage(String message) {
        send(ClientToServerProtocol.gameMessageString, message.getBytes());
    }

    private void send(ClientToServerProtocol messageType, byte[] message) {
        System.out.println("SEND: " + message.length);
        message = Util.combineArrays(new byte[]{messageType.messageNum}, message);
        io.getOutgoingConnection().sendGameMessage(message);
    }

    private void gameStarted(byte[] message) {
        try {
            GameOptions options = GameOptions.getGameOptions(message, (short) 1);
            game.gameStarted(me, options);
        } catch (InvalidMessageException e) {
            return;
        }
    }

    public void leaveGame() {
        io.getOutgoingConnection().leaveGame();
    }

    public UserList getUserList() {
        return userList;
    }

    public ChatBox getChatBox() {
        return chat;
    }

    @Override
    public void addUser(User user) {
        if (user.getUsername().equals(me.getUsername())) {
            return;
        }
        game.playerJoined(new Player(user.getUsername()));
    }

    @Override
    public void removeUser(String user) {
        game.playerLeft(user);
    }

    @Override
    public void renameUser(String from, String to) {
        System.out.println("STUB TODO: Rename User");
    }

    @Override
    public void loggedIn() {
        return;
    }

    @Override
    public void loggedOut() {
        return;
    }

}
