/*    
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.tbgf.server;

import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sf.goodgame.server.game.GameResult;
import net.sf.goodgame.server.game.GameServer;
import net.sf.goodgame.server.game.GameServerListener;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.games.GameOptions;
import net.sf.goodgame.shared.games.Player;
import net.sf.goodgame.shared.network.InvalidMessageException;
import net.sf.goodgame.tbgf.shared.ClientToServerProtocol;
import net.sf.goodgame.tbgf.shared.ServerToClientProtocol;

public abstract class TurnBasedGameServer implements GameServer {

    private boolean gameIsSetup;
    protected TurnBasedGame game;
    protected GameServerListener serverListener;
    protected Vector<Player> players;
    protected Map<Player, GameOptions> options;
    protected GameOptions globalGameOptions;
    protected boolean gameOver;
    protected Timer timeoutTimer;
    protected TimeoutTask timeoutTask;

    protected abstract GameOptions getSetupGameOptions();

    protected abstract void setupGame(GameOptions setupOptions);
    
    /**
     * @return the time (in milliseconds) required to declare a player as timed out.
     */
    protected abstract long getTimeoutLength();

    public TurnBasedGameServer() {
        options = new ConcurrentHashMap<Player, GameOptions>();
        players = new Vector<Player>();
        timeoutTask = new TimeoutTask();
        timeoutTimer = new Timer();
        globalGameOptions = new GameOptions();
    }

    public void messageFromUser(Player player, String message) {
        System.out.println("Invalid Message: message came in from string");
        return;//Shouldn't ever happen.
    }

    public void messageFromUser(Player player, byte[] message) {
        ClientToServerProtocol messageType = ClientToServerProtocol.valueOf(message[0]);
        switch (messageType) {
            case setupGame:
                setupGame(message,(short) 1);
                break;
            case gameMessageBinary:
                playerMove(player, message, (short) 1);
                break;
            case gameMessageString:
                playerMove(player, new String(Util.extractArray(message, 1, message.length - 1)));
                break;
            default:
                System.out.println("Error: Invalid message in TBGF");
                break;
        }
    }

    private void setupGame(byte[] message,short read) {
        try {
            GameOptions setupOptions = GameOptions.getGameOptions(message, read);
            setupGame(setupOptions);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(TurnBasedGameServer.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    @Override
    public boolean canPlayerJoin() {
        return !gameOver;
    }

    @Override
    public void playerJoined(Player player) {

        players.add(player);
        options.put(player, new GameOptions());
        if(gameIsSetup)
        {
            game.playerJoined(player);
        }else
        {
            this.sendToOne(ServerToClientProtocol.setupGameOptions, getSetupGameOptions().getNetworkMessage(), player);
        }
    }

    public void playerLeft(Player player) {
        players.remove(player);
        options.remove(player);
        game.playerLeft(player);
    }

    public void setGameOptions(GameOptions myOptions) {
        globalGameOptions.putAll(myOptions);
    }

    public void setGameOptions(GameOptions options, Player p) {
        this.options.get(p).putAll(options);
    }

    public GameOptions getGameOptions() {
        return globalGameOptions;
    }

    public void sendGameOptions(Player to) {
        GameOptions myOptions = new GameOptions();
        myOptions.putAll(globalGameOptions); //Global Game Options are added first to ensure any player specific options override the globals.
        myOptions.putAll(options.get(to));
        sendToOne(ServerToClientProtocol.gameOptions, myOptions.getNetworkMessage(), to);
    }

    public void setServerGameListener(GameServerListener l) {
        serverListener = l;
    }

    public void sendMessage(byte[] message) {
        sendToAll(ServerToClientProtocol.gameMessageBinary, message);
    }

    public void sendMessage(String message) {
        sendToAll(ServerToClientProtocol.gameMessageString, message.getBytes());
    }

    public void sendMessageTo(byte[] message, Player player) {
        this.sendToOne(ServerToClientProtocol.gameMessageBinary, message, player);
    }

    public void sendMessageTo(String message, Player player) {
        sendToOne(ServerToClientProtocol.gameMessageString, message.getBytes(), player);
    }

    public void gameStarted() {
        this.serverListener.gameStarted();
        for (Player p : players) {

            GameOptions cur = new GameOptions();
            cur.putAll(globalGameOptions);
            cur.putAll(options.get(p));
            sendToOne(ServerToClientProtocol.gameStarted, cur.getNetworkMessage(), p);
        }
        //Start the timer.
        timeoutTimer.schedule(timeoutTask, getTimeoutLength());
    }

    public void gameOver(GameResult r, String message) {
        this.serverListener.gameOver();
        byte[] messageBytes = (message == null) ? new byte[0] : message.getBytes();
        gameOver = true;
        sendToAll(ServerToClientProtocol.gameOverString, messageBytes);
        gameResults(r);
    }

    protected void gameIsSetup()
    {
        
        serverListener.gameIsSetup();
        gameIsSetup = true;
    }
    public void sendGameIsSetup(Player p)
    {
        sendToOne(ServerToClientProtocol.gameIsSetup,getGameOptions().getNetworkMessage(),p);
    }
    public void gameOver(GameResult r, byte[] message) {
        if (message == null) {
            message = new byte[0];
        }
        gameOver = true;
        sendToAll(ServerToClientProtocol.gameOverBinary, message);
        gameResults(r);
    }

    private void gameResults(GameResult r) {
        if (r != null) {
            serverListener.gameResults(r);
        }
    }

    private void sendToAll(ServerToClientProtocol messageType, byte[] message) {
        message = Util.combineArrays(
                new byte[]{messageType.messageNum}, message);
        serverListener.messageFromGame(message, players.toArray(new Player[0]));
    }

    private void sendToOne(ServerToClientProtocol messageType, byte[] message, Player player) {
        byte[] networkMessage = Util.combineArrays(
                new byte[]{messageType.messageNum}, message);
        serverListener.messageFromGame(networkMessage, new Player[]{player});
    }

    private void playerMove(Player player, byte[] message, short read) {
        game.gameMessage(player, message, read);
    }

    private void playerMove(Player player, String message) {
        game.gameMessage(player, message);
    }

    private void sendTimeout() {
        sendToOne(ServerToClientProtocol.moveTimeout, new byte[0], game.getCurrentTurn());
    }

    public void turnOver(Player nextTurn) {
        byte[] message = nextTurn.getUsername().getBytes();
        sendToAll(ServerToClientProtocol.turnOver, message);
        //Start the timer over again
        try {
            timeoutTask.cancel();
            timeoutTask = new TimeoutTask();
            timeoutTimer.schedule(timeoutTask, this.getTimeoutLength());
        } catch (IllegalStateException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void destroy() {
        try {
            System.out.println("Destroy");
            timeoutTask.cancel();
            timeoutTimer.cancel();
            game = null;
        } catch (Exception e) {
            //We might get an exception here from either timeoutTask or timeoutTimer.
            System.out.println("ERR: " + e.getLocalizedMessage());
        }
    }

    protected class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            //Send Move Timout to the offending player
            sendTimeout();
            game.moveTimeout();
        }
    }
}
