/*
(C) 2008 Michael Chaten
This file is part of Good Game.
Good Game is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License
Good Game is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.client.networking;

import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import net.sf.goodgame.client.game.Game;
import net.sf.goodgame.client.gametype.GameType;
import net.sf.goodgame.client.gui.GoodGameUI;
import net.sf.goodgame.shared.network.ClientToServerHeader;
import net.sf.goodgame.shared.network.NetworkMessage;

/**
 *
 * @author Mike
 */
public class ClientToServerLogic implements OutgoingConnection {

    private final InputOutput io;
    private final GoodGameUI gui;
    private final Executor executor;

    public ClientToServerLogic(InputOutput io, GoodGameUI gui) {
        this.io = io;
        this.gui = gui;
        executor = Executors.newSingleThreadExecutor();
    }

    public void connect() {
        executor.execute(new Runnable() {

            public void run() {
                io.connect();
            }
        });
    }

    public void switchGameType(final GameType gt) {
        executor.execute(new Runnable() {

            public void run() {
                doSwitchGameType(gt);
            }
        });
    }

    public void handshake(final short version) {
        executor.execute(new Runnable() {

            public void run() {
                doHandshake(version);
            }
        });
    }

    private void doHandshake(short version) {
        NetworkMessage message = NetworkMessage.createMessage(ClientToServerHeader.handshake, version);
        io.send(message);
    }

    private void doSwitchGameType(GameType gt) {
        if(gt != null)
        {
            NetworkMessage message = NetworkMessage.createMessage(ClientToServerHeader.switchGameType, gt.ID);
            io.send(message);
        }
    }

    public void joinGame(final Game game) {
        executor.execute(new Runnable() {

            public void run() {
                doJoinGame(game);
            }
        });
    }

    private void doJoinGame(Game game) {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.joinGame, game.gameID);
        io.send(networkMessage);
    }

    public void leaveGame() {
        executor.execute(new Runnable() {

            public void run() {
                doLeaveGame();
            }
        });
    }

    public void doLeaveGame() {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.leaveGame, new byte[0]);
        io.send(networkMessage);
    }

    public void sendPrivateMessage(final String username, final String message) {
        executor.execute(new Runnable() {

            public void run() {
                doSendPrivateMessage(username, message);
            }
        });
    }

    private void doSendPrivateMessage(String username, String message) {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.privateMessage, username, message);
        io.send(networkMessage);
    }

    public void sendChat(final String message) {
        executor.execute(new Runnable() {

            public void run() {
                doSendChat(message);
            }
        });
    }

    private void doSendChat(String message) {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.chat, message.trim());
        if (networkMessage.getPayload().length <= 512) //Max allowable length of a chat message
        {
            io.send(networkMessage);
        }
    }

    public void register(final String username, final String password, final String email) {
        executor.execute(new Runnable() {

            public void run() {
                doRegister(username, password, email);
            }
        });
    }
    private static final byte MAX_USERNAME_LEN = 26;
    private static final byte MAX_PASSWORD_LEN = 26;

    private void doRegister(String username, String password, String email) {
        if (username.length() < MAX_USERNAME_LEN && password.length() < MAX_PASSWORD_LEN) {
            NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.register, username, password, email);
            io.encryptAndSend(networkMessage);
        }
    }

    public void login(final String username, final String password) {
        executor.execute(new Runnable() {

            public void run() {
                doLogin(username, password);
            }
        });
    }

    private void doLogin(String username, String password) {
        if (username.length() < MAX_USERNAME_LEN && password.length() < MAX_PASSWORD_LEN) {
            NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.login, username, password);
            io.encryptAndSend(networkMessage);
        } else {
            //Simulate a loginFailed message
            gui.showError("Username or password is too long");
        }
    }

    public void getUserInfo(final String username) {
        executor.execute(new Runnable() {

            public void run() {
                doGetUserInfo(username);
            }
        });
    }

    public void doGetUserInfo(String username) {
        if (username.length() < MAX_USERNAME_LEN) {
            NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.getUserInformation, username);
            io.send(networkMessage);
        }
    }

    public void logout() {
        executor.execute(new Runnable() {

            public void run() {
                doLogout();
            }
        });
    }

    public void doLogout() {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.logout, new byte[0]);
        io.send(networkMessage);
    }

    public void switchLobby(final short id) {
        executor.execute(new Runnable() {

            public void run() {
                doSwitchLobby(id);
            }
        });
    }

    public void doSwitchLobby(short id) {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.switchLobby, id);
        io.send(networkMessage);
    }

    public void createGame() {
        executor.execute(new Runnable() {

            public void run() {
                doCreateGame();
            }
        });
    }

    private void doCreateGame() {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.createGame);
        io.send(networkMessage);
    }
    public void sendGameMessage(final byte[] message) {
        executor.execute(new Runnable() {

            public void run() {
                doSendGameMessage(message);
            }
        });
    }

    public void sendGameMessage(final String message) {
        executor.execute(new Runnable() {

            public void run() {
                doSendGameMessage(message);
            }
        });
    }

    private void doSendGameMessage(byte[] message) {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.gameBytes, message);
        io.send(networkMessage);
    }

    private void doSendGameMessage(String message) {
        NetworkMessage networkMessage = NetworkMessage.createMessage(ClientToServerHeader.gameMessage, message);
        io.send(networkMessage);
    }

    public void addFriend(final String friend) {
        executor.execute(new Runnable()
        {
            public void run()
            {
                doAddFriend(friend);
            }
        });
    }
    private void doAddFriend(String friend)
    {
        NetworkMessage message = NetworkMessage.createMessage(ClientToServerHeader.addFriend,friend);
        io.send(message);
    }

    public void removeFriend(final String friend) {
        executor.execute(new Runnable()
        {
            public void run()
            {
                doRemoveFriend(friend);
            }
        });
    }
    private void doRemoveFriend(String friend)
    {
        NetworkMessage message = 
                NetworkMessage.createMessage(ClientToServerHeader.removeFriend, friend);
        io.send(message);
    }
    
}

