/*
(C) 2008 Michael Chaten
This file is part of Good Game.
Good Game is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License
Good Game is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.client.networking;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sf.goodgame.client.gui.GoodGameUI;
import net.sf.goodgame.shared.network.ClientToServerHeader;
import net.sf.goodgame.shared.Logging;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.network.ImmutableNetworkMessage;
import net.sf.goodgame.shared.network.InvalidMessageException;
import net.sf.goodgame.shared.network.ServerToClientHeader;
import net.sf.goodgame.shared.network.NetworkMessage;

/**
 *
 * @author Mike
 */
public class InputOutput implements Runnable
{
    private SocketChannel socket;
    private volatile boolean running;
    private final ClientToServerLogic clientLogic;
    private final ServerToClientLogic serverLogic;
    private final GoodGameUI ui;
    private ByteBuffer headerBuffer;
    private final InetSocketAddress serverAddress;
    public InputOutput(InetSocketAddress address, GoodGameUI ui)
    {
        this.ui = ui;
        headerBuffer = ByteBuffer.allocateDirect(ServerToClientHeader.HEADER_SIZE);
        serverAddress = address;
        this.clientLogic = new ClientToServerLogic(this, ui);
        this.serverLogic = new ServerToClientLogic(this, ui);
    }

    public OutgoingConnection getOutgoingConnection()
    {
        return clientLogic;
    }
    public IncomingConnection getIncomingConnection()
    {
        return serverLogic;
    }
    public void disconnect()
    {
        running = false;
        try {
            socket.close();
        } catch (IOException ex) {
        }
    }

    public void connect()
    {
        try
        {
           
            socket = SocketChannel.open();
            socket.configureBlocking(false);
            socket.connect(serverAddress);
            for (int i = 0; i < 10; i++)
            {
                if (socket.finishConnect())
                {
                    new Thread(this).start();
                     ui.connecting("Connected");
                     ui.connecting("Sending Handshake");
                     clientLogic.handshake(ClientToServerHeader.version);
                     return;
                }
                else
                {
                    try
                    {
                        synchronized(this) //We may get an java.lang.IllegalMonitorStateException without synchronized.
                        {
                            /* Min of 10 seconds wait for the 
                             connection to finish up. (10 loops * 1000ms = 10s)*/
                            this.wait(1000); 
                        }
                    }catch(InterruptedException e)
                    {}
                }
            }
            ui.disconnected();
            socket = null;
        } catch (IOException e)
        {
            ui.disconnected();
            socket = null;
        }
    }

    public void run()
    {
        running = true;
        Selector selector;
        try
        {
            selector = Selector.open();
            socket.register(selector,SelectionKey.OP_READ);
        } catch (IOException e)
        {
            e.printStackTrace();
            return;
        }
        while (running)
        {
            try
            {
                if (selector.select(100) != 0)
                {
                    Iterator iterator = selector.selectedKeys().iterator();
                    while (iterator.hasNext())
                    {
                        SelectionKey selectionKey = (SelectionKey) iterator.next();
                        if (selectionKey.isReadable())
                        {
                            SocketChannel socketChannel = (SocketChannel) selectionKey.channel();
                            try
                            {
                                analyzeMessage(socketChannel);
                            } catch (IOException e)
                            {
                                //According to my tests, this means that the channel died
                                //on the other end, but this end doesnt know about it
                                // Is It even possible to send 0 bytes???
                                Logging.log("Closing channel: " + e.getMessage(), Level.INFO);
                                selectionKey.cancel();
                                ui.disconnected();
                                return;
                            }
                        }
                        iterator.remove();
                    }
                }
            } catch (IOException e)
            {
                e.printStackTrace();
                Logging.log("Unknown socket error", Level.SEVERE);
                ui.disconnected();
                return;
            }
        }

        try
        {
            socket.close();
            ui.disconnected();
        } catch (IOException e)
        {
            //Do Nothing
        }

    }

    public void analyzeMessage(SocketChannel socket) throws IOException
    {
        Logging.log("Reading a message");

        //Read The Header
        socket.read(headerBuffer);
        if (headerBuffer.position() == 0)
        {
            throw new IOException("Read 0 bytes");
        }
        headerBuffer.flip();
        short messageNumber = headerBuffer.getShort();
        short messageLen = headerBuffer.getShort();
        headerBuffer.clear();
        //Read the message payload
        if (messageLen < 0)
        {
            return;
        }
        ByteBuffer messageBuffer = ByteBuffer.allocate(messageLen);
        int read = 0;
        int counter = 0;
        while (read != messageLen && counter < 100) //Handle Partial Reads 
        {
            read += socket.read(messageBuffer);
            counter++;
        }
        messageBuffer.flip();
        byte[] payload = messageBuffer.array();

        ServerToClientHeader messageType = ServerToClientHeader.getHeader(messageNumber);
        
        NetworkMessage networkMessage = new ImmutableNetworkMessage(messageType,null,payload);
        Logging.log( "Message Type: " + messageType + '\n');

        switch (messageType)
        {
            //case 0:break;//Can't Handshake if already initialized
            case handshake:
                serverLogic.handshake(networkMessage);
                break;
            case oldVersion:
                serverLogic.oldVersion(networkMessage);
                break;
            case loginOk:
                serverLogic.loginOk(networkMessage);
                break;
            case loginFailed:
                serverLogic.loginFailed(networkMessage);
                break;
            case logout:
                serverLogic.logout(networkMessage);
                break;
            case switchLobby:
                serverLogic.switchLobby(networkMessage);
                break;
            case addLobbies:
                serverLogic.addLobbies(networkMessage);
                break;
            case modifyLobbies:
                serverLogic.modifyLobbies(networkMessage);
                break;
            case removeLobbies:
                serverLogic.removeLobbies(networkMessage);
                break;
            case gameMessage:
                serverLogic.gameMessage(networkMessage);
                break;
            case gameBytes:
                serverLogic.gameBytes(networkMessage);
                break;
            case chat:
                serverLogic.chat(networkMessage);
                break;
            case privateMessage:
                serverLogic.privateMessage(networkMessage);
                break;
            case addGames:
                serverLogic.addGames(networkMessage);
                break;
            case modifyGames:
                serverLogic.modifyGames(networkMessage);
                break;
            case removeGames:
                serverLogic.removeGames(networkMessage);
                break;
            case joinGame:
                serverLogic.joinGame(networkMessage);
                break;
            case leaveGame:
                serverLogic.leaveGame(networkMessage);
                break;
            case createGame:
                //serverLogic.createGame();
                break;
            case loadGame:
                //serverLogic.loadGame();
                break;
            case addUsers:
                serverLogic.addUsers(networkMessage);
                break;
            case modifyUsers:
                serverLogic.modifyUsers(networkMessage);
                break;
            case removeUsers:
                serverLogic.removeUsers(networkMessage);
                break;
            case addFriend:
                serverLogic.addFriend(networkMessage);
                break;
            case removeFriend:
                serverLogic.removeFriend(networkMessage);
                break;
            case friendLoggedIn:
                serverLogic.friendLoggedIn(networkMessage);
                break;
            case friendLoggedOut:
                serverLogic.friendLoggedOut(networkMessage);
                break;
            case userInformation:
                serverLogic.userInformation(networkMessage);
                break;
            case addGameTypes:
                serverLogic.addGameTypes(networkMessage);
                break;
            case modifyGameTypes:
                //serverLogic.modifyGameTypes();
                break;
            case removeGameTypes:
                //serverLogic.removeGameTypes();
                break;
            case switchGameType:
                serverLogic.switchGameType(networkMessage);
                break;
            case serverMessage:
                serverLogic.serverMessage(networkMessage);
                break;
            default:
                System.out.println("Unkown message type:" +messageType);
                return;
        }
    }

    public void send(NetworkMessage message)
    {
        if (message != null)
        {
            System.out.println("Sending " + message);
            ByteBuffer[] writeBuffers = new ByteBuffer[]
            {
                ByteBuffer.wrap(message.getHeader()),ByteBuffer.wrap(message.getPayload())
            };
            try
            {
                short headerMessageLen = Util.byteArrayToShort(message.getHeader(), 2);
                System.out.println("Header has a len of: "+headerMessageLen);
            }catch(InvalidMessageException e)
            {
                e.printStackTrace();
            }
            long bytesWritten = 0;
            long bytesToWrite = writeBuffers[0].limit() + writeBuffers[1].limit();
            while (bytesWritten < bytesToWrite && socket.isOpen())
            {
                try
                {
                    bytesWritten += socket.write(writeBuffers);
                } catch (IOException e)
                {
                    //Logging.log("Exception while writing to: " + user.getUsername());
                    e.printStackTrace();
                    disconnect();
                    break;
                }
            }
        }
    }

    public void encryptAndSend(NetworkMessage message)
    {

        if (Util.canEncrypt())
        {
            try
            {
                message.encrypt();
                send(message);
            } catch (Exception e)
            {
                e.printStackTrace();
                return;
            }
        }

    }

}
