/*
(C) 2008 Michael Chaten
This file is part of Good Game.
Good Game is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License
Good Game is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.client.networking;

import java.security.KeyFactory;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.X509EncodedKeySpec;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sf.goodgame.client.user.User;
import net.sf.goodgame.client.game.Game;
import net.sf.goodgame.client.game.GameClient;
import net.sf.goodgame.client.game.GameList;
import net.sf.goodgame.client.gametype.GameType;
import net.sf.goodgame.client.gametype.GameTypeList;
import net.sf.goodgame.client.gui.chat.ChatListener;
import net.sf.goodgame.client.gui.GoodGameUI;
import net.sf.goodgame.client.gui.Lobby;
import net.sf.goodgame.client.gui.LobbyList;
import net.sf.goodgame.client.gui.chat.PrivateChatController;
import net.sf.goodgame.client.user.friends.FriendsListener;
import net.sf.goodgame.client.user.UserList;
import net.sf.goodgame.shared.UserInfoElement;
import net.sf.goodgame.shared.games.GameState;
import net.sf.goodgame.shared.UserType;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.games.table.GameTableLayout;
import net.sf.goodgame.shared.games.table.GameTableRow;
import net.sf.goodgame.shared.network.InvalidMessageException;
import net.sf.goodgame.shared.network.NetworkMessage;

/**
 *
 * @author Mike
 */
public class ServerToClientLogic implements IncomingConnection {

    private final InputOutput io;
    private GoodGameUI ui;
    private UserList userList;
    private ChatListener chatListener;
    private GameList gameListener;
    private GameClient game;
    private GameTypeList gameTypeList;
    private PrivateChatController privateChatController;
    private LobbyList lobbyList;
    private FriendsListener friendsListener;

    public ServerToClientLogic(InputOutput io, GoodGameUI ui) {
        this.io = io;
        this.ui = ui;
    }

    @Override
    public void setUserList(UserList ul) {
        userList = ul;
    }

    @Override
    public void setChatListener(ChatListener cl) {
        chatListener = cl;
    }

    @Override
    public void setGameList(GameList gl) {
        gameListener = gl;
    }

    @Override
    public void setGameClient(GameClient g) {
        game = g;
    }

    @Override
    public void setGameTypeList(GameTypeList gt) {
        gameTypeList = gt;
    }

    @Override
    public void setPrivateChatController(PrivateChatController controller) {
        privateChatController = controller;
    }

    @Override
    public void setFriendsListener(FriendsListener listener) {
        friendsListener = listener;
    }

    @Override
    public void setUI(GoodGameUI ui) {
        this.ui = ui;
    }

    @Override
    public void setLobbyList(LobbyList list) {
        lobbyList = list;
    }

    public void handshake(NetworkMessage networkMessage) {

        try {
            RSAPublicKey pubKey;
            Object[] data = networkMessage.decodeMessage(false, new byte[0].getClass());
            pubKey = (RSAPublicKey) KeyFactory.getInstance("RSA").
                    generatePublic(new X509EncodedKeySpec((byte[]) data[0]));
            Util.setKeys(pubKey, null);
            ui.connecting("Handshake complete");
            ui.connected();
        } catch (Exception e) {
            ui.connecting("Cannot complete handshake, please update your version of java at java.sun.com");
            io.disconnect();
            ui.disconnected();
            e.printStackTrace();
            return;
        }

    }

    public void oldVersion(NetworkMessage networkMessage) {
        try {

            ui.connecting("Please update your version of Good Game to version: " + (Short) networkMessage.decodeMessage(false, Short.class)[0]);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void loginOk(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, UserType.class, String.class);
            UserType type = (UserType) data[0];
            String name = (String) data[1];
            System.out.println("Name is: " + name);
            ui.login(name, type);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void loginFailed(NetworkMessage networkMessage) {
        ui.showError("Invalid Username or Password");
    }

    public void logout(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, String.class);
            ui.login((String) data[0], UserType.guest);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void switchLobby(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, Short.class);
            ui.switchLobby((Short) data[0]);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void serverMessage(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, String.class);
            ui.serverMessage((String) data[0]);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void switchGameType(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, Short.class);
            short gameTypeID = (Short) data[0];
            System.out.println("Switching to gametype : " + gameTypeID);
            ui.switchGameType(gameTypeList.getGameType(gameTypeID));
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void addLobbies(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, Short.class, String.class);
            for (int i = 0; i < data.length; i += 2) {
                short lobbyID = (Short) data[i];
                String lobbyName = (String) data[i + 1];
                lobbyList.addLobby(new Lobby(lobbyID, lobbyName));
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void modifyLobbies(NetworkMessage networkMessage) {
        throw new UnsupportedOperationException("Unsupported Operation");
    }

    public void removeLobbies(NetworkMessage networkMessage) {
        throw new UnsupportedOperationException("Unsupported Operation");
    }

    public void gameMessage(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, String.class);
            game.gameMessage((String) data[0]);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void gameBytes(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, new byte[0].getClass());
            game.gameMessage((byte[]) data[0]);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void chat(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, String.class, String.class);
            String name = (String) data[0];
            String message = (String) data[1];
            chatListener.incomingChat(name, message);
        } catch (InvalidMessageException ex) {
        }
    }

    public void privateMessage(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, String.class, String.class);
            String name = (String) data[0];
            String message = (String) data[1];
            privateChatController.privateMessage(name, message);
        } catch (InvalidMessageException ex) {
        }
    }

    public void addGames(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, Short.class, Short.class, GameState.class, GameTableRow.class);
            for (int i = 0; i < data.length;) {
                short gameID = (Short) data[i++];
                short gameTypeID = (Short) data[i++];
                GameType gameType = gameTypeList.getGameType(gameTypeID);
                GameState gameState = (GameState) data[i++];
                GameTableRow row = (GameTableRow) data[i++];

                Game newGame = new Game(gameID, gameType, gameState, row);
                gameListener.addGame(newGame);
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void modifyGames(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, Short.class, GameState.class, GameTableRow.class);
            for (int i = 0; i < data.length;) {
                short gameID = (Short) data[i++];
                GameState gameState = (GameState) data[i++];
                GameTableRow row = (GameTableRow) data[i++];
                Game updatedGame = Game.getGame(gameID);
                updatedGame.setRow(row);
                updatedGame.setState(gameState);
                gameListener.updateGame(updatedGame);
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void removeGames(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, Short.class);
            for (int i = 0; i < data.length;) {
                short gameID = (Short) data[i++];
                gameListener.removeGame(Game.getGame(gameID));
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void joinGame(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, Short.class);
            short gameID = (Short) data[0];
            GameType gt = gameTypeList.getGameType(gameID);
            ui.joinGame(gt);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void leaveGame(NetworkMessage networkMessage) {
        ui.leaveGame();
    }

    public void addUsers(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, UserType.class, String.class);
            for (int i = 0; i < data.length;) {
                UserType type = (UserType) data[i++];
                String name = (String) data[i++];
                User user = new User(name, type);
                System.out.println("Add User: "+user.getUsername());
                userList.addUser(user);
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void modifyUsers(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, UserType.class, String.class, String.class);
            for (int i = 0; i < data.length;) {
                UserType type = (UserType) data[i++];
                String newName = (String) data[i++];
                String oldName = (String) data[i++];
                userList.renameUser(oldName, newName, type);
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void removeUsers(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, String.class);
            for (int i = 0; i < data.length;) {
                String name = (String) data[i++];
                System.out.println("Remove user: "+name);
                userList.removeUser(name);
            }
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    //TODO: Implement
    public void userInformation(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(false, String.class, UserInfoElement[].class);
            String username = (String) data[0];
            UserInfoElement[] elements = (UserInfoElement[]) data[1];
            ui.showUserInfo(username, Arrays.asList(elements));
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ServerToClientLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void addGameTypes(NetworkMessage networkMessage) {
        try {
            Object[] data = networkMessage.decodeMessage(true, Short.class, String.class, String.class, String.class, GameTableLayout.class);
            for (int i = 0; i < data.length;) {
                short ID = (Short) data[i++];
                String className = (String) data[i++];
                String jarName = (String) data[i++];
                String name = (String) data[i++];
                GameTableLayout layout = (GameTableLayout) data[i++];
                GameType gt = new GameType(name, jarName, className, ID, layout);
                gameTypeList.addGameType(gt);
            }
        } catch (InvalidMessageException e) {
        }
    }

    public void addFriend(NetworkMessage message) {
        if (friendsListener == null) {
            return;
        }
        try {
            Object[] data = message.decodeMessage(true, String.class);
            for (Object friend : data) {
                friendsListener.addFriend((String) friend);
            }
        } catch (InvalidMessageException e) {
        }
    }

    public void removeFriend(NetworkMessage message) {

        if (friendsListener == null) {
            return;
        }
        try {
            Object[] data = message.decodeMessage(false, String.class);
            friendsListener.removeFriend((String) data[0]);
        } catch (InvalidMessageException e) {
        }
    }

    public void friendLoggedIn(NetworkMessage message) {
        if (friendsListener == null) {
            return;
        }
        try {
            Object[] data = message.decodeMessage(false, String.class);
            friendsListener.friendLoggedIn((String) data[0]);
            chatListener.incomingChat("SERVER", data[0] + " has logged in");
        } catch (InvalidMessageException e) {
        }
    }

    public void friendLoggedOut(NetworkMessage message) {
        if (friendsListener == null) {
            return;
        }
        try {
            Object[] data = message.decodeMessage(false, String.class);
            friendsListener.friendLoggedOut((String) data[0]);
            chatListener.incomingChat("SERVER", data[0] + " has logged out");
        } catch (InvalidMessageException e) {
        }
    }
}
