/*    
(C) 2007,2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.server;

import java.util.Vector;
import java.util.concurrent.ConcurrentSkipListSet;
import net.sf.goodgame.server.user.User;
import net.sf.goodgame.shared.network.MutableNetworkMessage;
import net.sf.goodgame.shared.network.NetworkMessage;
import net.sf.goodgame.shared.network.ServerToClientHeader;

public class List<A> {

    private final ConcurrentSkipListSet<User> listeners;
    private final Vector<A> data;
    private final MutableNetworkMessage<A> initializeMessage;
    private volatile boolean enabled;
    private final ServerToClientHeader header;

    public List(ServerToClientHeader initializeHeader) {
        this(initializeHeader, true);
    }

    public List(ServerToClientHeader initializeHeader, boolean enabled) {

        data = new Vector<A>();
        listeners = new ConcurrentSkipListSet<User>();
        initializeMessage = new MutableNetworkMessage(initializeHeader);
        this.enabled = enabled;
        header = initializeHeader;
    }

    public void disable() {
        enabled = false;
    }

    public void enable() {
        enabled = true;
        if (data.size() == 0) {
            System.out.println("List.java:Data size == 0 for list: " + header);
            return;
        }
        synchronized (listeners) {
            synchronized (data) {
                for (User listener : listeners) {
                    initializeListener(listener);
                }
            }
        }
    }

    protected void addData(A add, NetworkMessage netMessage) {
        synchronized (data) {
            data.add(add);
            initializeMessage.put(add, netMessage);
        }
    }

    protected void updateData(A update,NetworkMessage netMessage) {
        synchronized (data) {
            initializeMessage.put(update,netMessage);
        }
    }

    protected void removeData(A remove) {
        synchronized (data) {
            data.remove(remove);
            initializeMessage.remove(remove);
        }
    }

    protected A[] getData(A[] array) {
        synchronized (data) {
            return data.toArray(array);
        }
    }

    public short getSize() {
        synchronized (data) {
            return (short) data.size();
        }
    }

    protected boolean contains(A data) {
        synchronized (data) {
            return this.data.contains(data);
        }
    }

    /**
     * If an extended class overrides ths method it must eventually call this method
     * @param listener
     */
    public final void addListener(User user) {
        synchronized (listeners) {
            listeners.add(user);
            initializeListener(user);
        }
    }

    /**
     * If an extended class overrides ths method it must eventually call this method
     * @param listener
     */
    public final void removeListener(User listener) {
        synchronized (listeners) {
            System.out.println("Removed "+listener+" ? "+listeners.remove(listener));
        }
    }

    final public User[] getListeners() {
        return listeners.toArray(new User[0]);
    }

    protected void initializeListener(User listener) {
        if (enabled) {
            synchronized (data) {
                listener.send(initializeMessage);
            }
        }
    }

    protected void updateListeners(NetworkMessage message) {
        if (enabled) {
            synchronized (listeners)//Protect against ConcurrentModificationExceptions
            {
                for (User listener : listeners) {
                    listener.send(message);
                }
            }
        }
    }

    final public boolean hasListener(User listener) {
        return listeners.contains(listener);
    }
}
