/*    
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.server.game;

import java.util.HashMap;
import java.util.Map;
import net.sf.goodgame.server.lobby.GameList;
import net.sf.goodgame.server.lobby.UserList;
import net.sf.goodgame.server.networking.ServerToClientLogic;
import net.sf.goodgame.server.persistence.DatabaseManager;
import net.sf.goodgame.server.user.Results;
import net.sf.goodgame.server.user.User;
import net.sf.goodgame.shared.games.GameState;
import net.sf.goodgame.shared.games.Player;
import net.sf.goodgame.shared.games.table.GameTableRow;
import net.sf.goodgame.shared.network.NetworkMessage;

public class Game implements GameServerListener {

    private final UserList userList;
    public final short ID;
    public final GameType gameType;
    private final GameList gameList;
    private final Map<User, Player> playersMap;
    private final Map<Player, User> usersMap;
    private GameServer game;
    private GameState gameState;

    public Game(GameList gameList, short ID, GameType gt) {
        this.gameList = gameList;
        userList = new UserList();
        this.ID = ID;
        this.gameType = gt;
        gameState = GameState.settingUp;
        playersMap = new HashMap<User, Player>();
        usersMap = new HashMap<Player, User>();
        game = gameType.createNewGameServer();
        game.setServerGameListener(this);
    }

    public boolean addUser(User user) {
        if (game.canPlayerJoin()) {
            Player player = new Player(user.getUsername());
            user.send(ServerToClientLogic.joinGame(this));
            userList.addUser(user);
            userList.addListener(user);
            playersMap.put(user, player);
            usersMap.put(player, user);
            game.playerJoined(player);
            if (GameState.settingUp != gameState) {
                gameList.updateGame(this);
            }
            return true;
        }
        return false;
    }

    public void removeUser(User user) {
        userList.removeListener(user);
        userList.removeUser(user);
        Player removed = playersMap.remove(user);
        usersMap.remove(removed);
        game.playerLeft(removed);
        if (userList.getSize() == 0) {
            deleteGame();
        } else {
            gameList.updateGame(this);
        }
    }

    private void deleteGame() {
        for (User listener : userList.getListeners()) {
            listener.leaveGame();
        }
        gameList.removeGame(this);
        destroyGame();
    }

    public byte getCurrentPlayers() {
        return (byte) userList.getSize();
    }

    public void sendChat(User from, String message) {
        userList.sendChat(from, message);
    }

    @Override
    public void messageFromGame(String message, Player[] towho) {
        NetworkMessage networkMessage = ServerToClientLogic.gameMessage(message);
        for (Player player : towho) {
            User user = User.getUser(player.getUsername());
            if (user != null && userList.hasUser(user)) {
                user.send(networkMessage);
            }
        }
    }

    @Override
    public void messageFromGame(byte[] message, Player[] towho) {
        NetworkMessage networkMessage = ServerToClientLogic.gameBytes(message);
        for (Player player : towho) {
            User user = usersMap.get(player);
            if (user != null) {
                user.send(networkMessage);
            }
        }
    }

    private void destroyGame() {
        game.destroy();
        game = null;
    }

    @Override
    public void gameResults(GameResult result) {
        DatabaseManager.getDatabase().storeResults(gameType, result);
        if (result.countsTowardsStats()) {
            if (result.wasTied()) {
                for (User u : result.getTie()) {
                    Results r = DatabaseManager.getDatabase().getResults(gameType, u);
                    System.out.println("Results for " + u + ": " + r);
                }
            } else {
                for (User u : result.getWinners()) {
                    Results r = DatabaseManager.getDatabase().getResults(gameType, u);
                    System.out.println("Results for " + u + ": " + r);
                }
                for (User u : result.getLosers()) {
                    Results r = DatabaseManager.getDatabase().getResults(gameType, u);
                    System.out.println("Results for " + u + ": " + r);
                }
            }
        }
    }

    public void messageToGame(User user, byte[] message) {
        Player player = playersMap.get(user);
        game.messageFromUser(player, message);
    }

    public void messageToGame(User user, String message) {
        Player player = playersMap.get(user);
        game.messageFromUser(player, message);
    }

    public void gameOver() {
        setGameState(GameState.finished);
    }

    private void setGameState(GameState state) {
        gameState = state;
        gameList.updateGame(this);
    }

    public void gameLoading() {
        setGameState(GameState.loading);
    }

    public void gameStarted() {
        setGameState(GameState.playing);
    }

    public GameState getState() {
        return gameState;
    }

    public void gameIsSetup() {
        //Don't use setGameState here because we can't update a game we haven't added yet
        gameState = GameState.canJoin;
        gameList.addGame(this);
    }

    public GameTableRow getGameTableRow() {
        return game.getGameTableRow();
    }
}
