/*    
    (C) 2008 Michael Chaten
 	This file is part of GoodGame.

    GoodGame is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 3 of the License

    GoodGame is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.goodgame.server.networking;

import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sf.goodgame.server.Logging;
import net.sf.goodgame.server.Processor;
import net.sf.goodgame.server.lobby.Lobby;
import net.sf.goodgame.server.game.Game;
import net.sf.goodgame.server.game.GameType;
import net.sf.goodgame.server.persistence.DatabaseManager;
import net.sf.goodgame.server.user.User;
import net.sf.goodgame.shared.UserInfoElement;
import net.sf.goodgame.shared.network.ClientToServerHeader;
import net.sf.goodgame.shared.UserType;
import net.sf.goodgame.shared.network.InvalidMessageException;
import net.sf.goodgame.shared.network.NetworkMessage;

/**
 *
 * @author Mike
 */
public class ClientToServerLogic {

    private final Processor processor;

    public ClientToServerLogic(Processor processor) {
        this.processor = processor;
    }

    public void handShake(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doHandShake(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doHandShake(User user, NetworkMessage payload) {
        try {
            Object[] data = payload.decodeMessage(false,Short.class);
            short version = (Short) data[0];
            if (version == ClientToServerHeader.version) {
                //Send the public key
                user.send(ServerToClientLogic.handShake());

                user.initialized();//The User is now initialized and may now do things outside of the handshake.
            } else {
                //Notify of an old version
                user.send(ServerToClientLogic.oldVersion());
                //Disconnect user;
                user.disconnect();
            }
        } catch (InvalidMessageException e) {
            user.send(ServerToClientLogic.oldVersion());
            user.disconnect();
        }
    }

    public void login(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doLogin(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doLogin(User user, NetworkMessage payload) {
        try {
            payload.decrypt();
            Object[] data = payload.decodeMessage(false,String.class, String.class);
            String username = (String) data[0];
            String password = (String) data[1];
            user.login(username, password);
        } catch (InvalidMessageException e) {
            //Log
            Logging.log("Recieved Invalid Message while logging in from user: " + user);
        }

    }

    public void logout(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doLogout(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doLogout(User user, NetworkMessage payload) {
        user.logout();
    }

    public void register(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doRegister(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doRegister(User user, NetworkMessage payload) {
        payload.getPayload();
        payload.decrypt();
        try {
            Object[] data = payload.decodeMessage(false,String.class, String.class, String.class);
            String username = (String) data[0];
            String password = (String) data[1];
            String email = (String) data[2];
            boolean result = DatabaseManager.getDatabase().registerUser(username, password, email, UserType.regular);
            if (result) {
                user.login(username, password);
            } else {
                user.send(ServerToClientLogic.loginFailed());
            }
        } catch (InvalidMessageException e) {
        }
    }

    public void joinGame(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doJoinGame(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doJoinGame(User user, NetworkMessage payload) {
        try {
            Object[] data = payload.decodeMessage(false,Short.class);
            short id = (Short) data[0];
            user.joinGame(id);
        } catch (InvalidMessageException e) {
        }
    }

    public void leaveGame(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doLeaveGame(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doLeaveGame(User user, NetworkMessage payload) {
        user.leaveGame();
    }
    /*
     * 2 bytes describing  the game type id. 
     * 1 Byte for the maximum number of players 
     * 1 Byte for the length of the name of the game, followed by the name of the game. 
     * 2 bytes for the length of the create game message followed by the create game message.
     * 
     */

    public void makeGame(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doMakeGame(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doMakeGame(User user, NetworkMessage payload) {
       user.makeGame();
    }

    public void chat(final User user, final NetworkMessage payload) {

        Runnable run = new Runnable() {

            public void run() {
                doChat(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doChat(User user, NetworkMessage payload) {
        try {
            Game game = user.getCurrentGame();
            Lobby lobby = user.getCurrentLobby();
            //Chat Message may only be 512 bytes or less.
            Object[] data = payload.decodeMessage(false,String.class);
            String message = (String) data[0];
            if (game != null) {
                game.sendChat(user, message);
            } else if (lobby != null) {
                lobby.sendChat(user, message);
            }
        } catch (InvalidMessageException e) {
        }
    }

    public void privateMessage(final User user, final NetworkMessage payload) {

        Runnable run = new Runnable() {

            public void run() {
                doPrivateMessage(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doPrivateMessage(User user, NetworkMessage payload) {
        try
        {
            Object[] data = payload.decodeMessage(false,String.class, String.class);
            String name = (String) data[0];
            String message = (String) data[1];
            User to = User.getUser(name);
            if (to == null) {
                return;
            }
            payload = ServerToClientLogic.privateMessage(user, message);
            to.send(payload);
            return;
        }catch(InvalidMessageException e)
        {
            
        }
    }

    public void switchLobby(final User user, final NetworkMessage payload) {

        Runnable run = new Runnable() {

            public void run() {
                doSwitchLobby(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doSwitchLobby(User user, NetworkMessage payload) {
        try
        {
            Object[] data = payload.decodeMessage(false,Short.class);
            short lobbyID = (Short) data[0];
            Lobby lobby = Lobby.getLobby(lobbyID);
            if (lobby != null && (user.getCurrentLobby() == null || lobby.ID != user.getCurrentLobby().ID)) {
                user.switchLobby(lobby);
            }
        }catch(InvalidMessageException e)
        {
            
        }
    }

    public void switchGameType(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doSwitchGameType(user, payload);
            }
        };
        processor.execute(user, run);
    }

    public void doSwitchGameType(final User user, final NetworkMessage payload) {
        try
        {
            Object[] data = payload.decodeMessage(false,Short.class);
            short gametypeID = (Short) data[0];
            GameType gametype = GameType.getGameType(gametypeID);
            user.switchGameType(gametype);
        }catch(InvalidMessageException e)
        {
            
        }
    }

    public void getUserInformation(final User user, final NetworkMessage payload) {

        Runnable run = new Runnable() {

            public void run() {
                doGetUserInformation(user, payload);
            }
        };
        processor.execute(user, run);
    }
    
    private void doGetUserInformation(User user, NetworkMessage payload) {
        try {
            Object[] data = payload.decodeMessage(false, String.class);
            NetworkMessage message = ServerToClientLogic.userInformation((String)data[0]);
            if(message == null)
                return;
            user.send(message);
        } catch (InvalidMessageException ex) {
            Logger.getLogger(ClientToServerLogic.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void setUserInformation(final User user, final NetworkMessage payload) {

        Runnable run = new Runnable() {

            public void run() {


                doSetUserInformation(user, payload);
            }
        };
        processor.execute(user, run);
    }
    //NOT IMPLEMENTED
    private void doSetUserInformation(User user, NetworkMessage payload) {
        //Todo: Implement
    }

    public void gameMessage(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doGameMessage(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doGameMessage(User user, NetworkMessage payload) {
        Game game = user.getCurrentGame();
        if (game != null) {
            try
            {
                Object[] data = payload.decodeMessage(false,String.class);
                String message = (String) data[0];
                game.messageToGame(user, message);
            }catch(InvalidMessageException e)
            {
                
            }
        }
    }

    public void gameBytes(final User user, final NetworkMessage payload) {
        Runnable run = new Runnable() {

            public void run() {
                doGameBytes(user, payload);
            }
        };
        processor.execute(user, run);
    }

    private void doGameBytes(User user, NetworkMessage payload) {
        Game game = user.getCurrentGame();
        if (game != null) {
            System.out.println("GameBytesLen: " +payload.getPayload().length);
            try
            {
                Object[] message = payload.decodeMessage(false, new byte[0].getClass());
                game.messageToGame(user,(byte[])message[0]);
            }catch(InvalidMessageException e)
            {
                System.out.println("Recieved invalid game bytes: "+e.getLocalizedMessage());
            }
        }
    }
  
    public void addFriend(final User user, final NetworkMessage payload)
    {
        Runnable run = new Runnable()
        {
            public void run()
            {
                doAddFriend(user,payload);
            }
        };
        processor.execute(user, run);
    }
    private void doAddFriend(User user, NetworkMessage payload)
    {
        try
        {
            Object[] data = payload.decodeMessage(false, String.class);
            user.addFriend((String)data[0]);
        }catch(InvalidMessageException e)
        {
            
        }
    }
    public void removeFriend(final User user, final NetworkMessage payload)
    {
        Runnable run = new Runnable()
        {
            public void run()
            {
                doRemoveFriend(user,payload);
            }
        };
        processor.execute(user, run);
        }
    private void doRemoveFriend(User user, NetworkMessage payload)
    {
          try
        {
            Object[] data = payload.decodeMessage(false, String.class);
            user.removeFriend((String)data[0]);
        }catch(InvalidMessageException e)
        {
            
        }
    }
}
