/*    
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.server.networking;

import java.util.logging.Logger;
import net.sf.goodgame.shared.Logging;
import net.sf.goodgame.server.user.User;
import net.sf.goodgame.server.*;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Level;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.network.ClientToServerHeader;
import net.sf.goodgame.shared.network.ImmutableNetworkMessage;
import net.sf.goodgame.shared.network.InvalidMessageException;
import net.sf.goodgame.shared.network.NetworkMessage;

/**
 *
 * @author Mike Chaten
 */
public class InputOutput implements Runnable {

    private final ConcurrentLinkedQueue<User> registrationQueue;
    private final ClientToServerLogic logic;
    private ConcurrentHashMap<SocketChannel, User> socketMap;
    private ByteBuffer headerBuffer;
    private Selector selector;
    private final AcceptConnections userConnections;
    private volatile boolean running;

    public InputOutput(int port, int backlog, Processor processor) {
        registrationQueue = new ConcurrentLinkedQueue<User>();
        logic = new ClientToServerLogic(processor);
        userConnections = new AcceptConnections(port, backlog);
        socketMap = new ConcurrentHashMap<SocketChannel, User>();
        running = false;
        headerBuffer = ByteBuffer.allocateDirect(ClientToServerHeader.HEADER_SIZE);
    }

    public void userDisconnected(User user) {
        socketMap.remove(user.socket);
        if (user.socket.isRegistered()) {
            user.socket.keyFor(selector).cancel();
        }
        try {
            user.socket.close();
        } catch (IOException e) {
            //Meh.
            e.printStackTrace();
        }
    }

    @Override
    public void run() {
        running = true;
        Thread acceptThread = new Thread(userConnections);
        acceptThread.start();

        try {
            selector = Selector.open();
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        while (running) {
            try {
                if (selector.select(100) != 0) {
                    Iterator iterator = selector.selectedKeys().iterator();
                    while (iterator.hasNext()) {
                        SelectionKey key = (SelectionKey) iterator.next();
                        if (key.isReadable()) {
                            SocketChannel socket = (SocketChannel) key.channel();
                            try {
                                analyzeMessage(socket);
                            } catch (IOException e) {
                                Logging.log("Closing channel: " + e.getMessage(), Level.INFO);
                                key.cancel();
                                socketMap.get(socket).disconnect();
                            }
                        }
                        iterator.remove();
                    }
                }
            } catch (IOException e) {
                e.printStackTrace();
                Logging.log("Unknown socket error", Level.SEVERE);
                return;
            }
            while (!registrationQueue.isEmpty()) {
                System.out.println("There are: " + registrationQueue.size() + " users waiting to finish connecting");
                User user = registrationQueue.poll();
                try {
                    user.socket.register(selector, SelectionKey.OP_READ);
                //lobbyServer.finishNewUser(session);
                } catch (IOException e) {
                    Logging.log(e.getMessage() + " caused by: " + e.getCause());
                    user.disconnect();
                }
            }
        }
    }

    public void stop() {
        running = false;
        System.out.println("Stoping the network");
        for(SelectionKey key: selector.keys()) {
            try {
                key.channel().close();
            } catch (IOException ex) {
                System.err.println("Exception while closing socket: "+ex.getMessage());
            }
        }
        try {
            selector.close();
        } catch (IOException ex) {
            System.err.println("Exception while closing the selector: "+ex.getMessage());
        }
        System.out.println("Sockets Closed");
    }

    private void analyzeMessage(SocketChannel socket) throws IOException {
        Logging.log("Reading a message");
        User user = socketMap.get(socket);
        if (user == null) {
            throw new IOException("Invalid user while analyzing");
        }
        //Read The Header
        socket.read(headerBuffer);
        if (headerBuffer.position() == 0) {
            throw new IOException("Read 0 bytes");
        }
        headerBuffer.flip();
        short messageNumber = headerBuffer.getShort();
        short messageLen = headerBuffer.getShort();
        headerBuffer.clear();
        //Read the message payload
        if (messageLen < 0) {
            return;
        }
        ByteBuffer messageBuffer = ByteBuffer.allocate(messageLen);
        int read = 0;
        int counter = 0;
        while (read != messageLen && counter < 100) //Handle Partial Reads 
        {
            read += socket.read(messageBuffer);
            counter++;
        }
        messageBuffer.flip();

        ClientToServerHeader messageType = ClientToServerHeader.getHeader(messageNumber);

        NetworkMessage message = new ImmutableNetworkMessage(messageType, null, messageBuffer.array());
        Logging.log("Message Type: " + messageType + '\n');

        if (user.isInitialized()) {
            switch (messageType) {
                //case 0:break;//Can't Handshake if already initialized
                case login:
                    logic.login(user, message);
                    break;
                case logout:
                    logic.logout(user, message);
                    break;
                case register:
                    logic.register(user, message);
                    break;
                case joinGame:
                    logic.joinGame(user, message);
                    break;
                case leaveGame:
                    logic.leaveGame(user, message);
                    break;
                case createGame:
                    logic.makeGame(user, message);
                    break;
                case chat:
                    logic.chat(user, message);
                    break;
                case privateMessage:
                    logic.privateMessage(user, message);
                    break;
                case switchLobby:
                    logic.switchLobby(user, message);
                    break;
                case switchGameType:
                    logic.switchGameType(user, message);
                    break;
                case getUserInformation:
                    logic.getUserInformation(user, message);
                    break;
                case setUserInformation:
                    logic.setUserInformation(user, message);
                    break;
                case gameMessage:
                    logic.gameMessage(user, message);
                    break;
                case gameBytes:
                    logic.gameBytes(user, message);
                    break;
                case addFriend:
                    logic.addFriend(user, message);
                    break;
                case removeFriend:
                    logic.removeFriend(user, message);
                    break;
                default:
                    return;
            }
        } else if (messageNumber == 0) {
            logic.handShake(user, message);
        }
    }

    public void send(NetworkMessage message, User user) {
        if (user != null && message != null) {
            System.out.println("Sending " + message + " to user " + user);
            SocketChannel sc = user.socket;
            ByteBuffer[] writeBuffers = new ByteBuffer[]{
                ByteBuffer.wrap(message.getHeader()),
                ByteBuffer.wrap(message.getPayload())
            };
            long bytesWritten = 0;
            long bytesToWrite = writeBuffers[0].limit() + writeBuffers[1].limit();
            while (bytesWritten < bytesToWrite && sc.isOpen()) {
                try {
                    bytesWritten += sc.write(writeBuffers);
                } catch (IOException e) {
                    Logging.log("Exception while writing to: " + user.getUsername());
                    e.printStackTrace();
                    user.disconnect();
                    break;
                }
            }
        }
    }

    private void newUser(SocketChannel socket) {
        User user = new User(socket);
        socketMap.put(socket, user);
        registrationQueue.add(user);

    }

    class AcceptConnections implements Runnable {

        ServerSocketChannel incoming;
        int port;
        int backlog;

        AcceptConnections(int port, int backlog) {
            this.port = port;
            this.backlog = backlog;
        }

        @Override
        public void run() {
            try {
                incoming = ServerSocketChannel.open();
                incoming.socket().setReuseAddress(true);
                incoming.configureBlocking(true);
                Logging.log("Binding to port: " + port, Level.INFO);
                incoming.socket().bind(new InetSocketAddress(port), backlog);

            } catch (IOException e) {
                e.printStackTrace();
                return;
            }

            while (running) {
                try {
                    SocketChannel inc = incoming.accept();
                    inc.configureBlocking(false);
                    newUser(inc);
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }
}
