/*
(C) 2008,2009 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.server.persistence;

import java.security.MessageDigest;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import net.sf.goodgame.server.game.GameResult;
import net.sf.goodgame.server.game.GameType;
import net.sf.goodgame.server.persistence.jpa.FriendEntity;
import net.sf.goodgame.server.persistence.jpa.FriendEntityID;
import net.sf.goodgame.server.persistence.jpa.FriendEntityJpaController;
import net.sf.goodgame.server.persistence.jpa.GameTypeEntity;
import net.sf.goodgame.server.persistence.jpa.GameTypeEntityJpaController;
import net.sf.goodgame.server.persistence.jpa.UserEntity;
import net.sf.goodgame.server.persistence.jpa.UserEntityJpaController;
import net.sf.goodgame.server.persistence.jpa.UserInfoElementEntity;
import net.sf.goodgame.server.persistence.jpa.UserInfoElementEntityID;
import net.sf.goodgame.server.persistence.jpa.UserInfoElementEntityJpaController;
import net.sf.goodgame.server.persistence.jpa.UserResultsEntity;
import net.sf.goodgame.server.persistence.jpa.UserResultsEntityID;
import net.sf.goodgame.server.persistence.jpa.UserResultsEntityJpaController;
import net.sf.goodgame.server.persistence.jpa.exceptions.NonexistentEntityException;
import net.sf.goodgame.server.persistence.jpa.exceptions.PreexistingEntityException;
import net.sf.goodgame.server.user.FriendsList;
import net.sf.goodgame.server.user.Results;
import net.sf.goodgame.server.user.User;
import net.sf.goodgame.shared.UserInfoElement;
import net.sf.goodgame.shared.UserInfoType;
import net.sf.goodgame.shared.UserType;
import net.sf.goodgame.shared.Util;

/**
 *
 * @author mchaten
 */
public class JPADatabase implements Database {

    private UserEntityJpaController users;
    private GameTypeEntityJpaController gameTypes;
    private UserResultsEntityJpaController userResults;
    private FriendEntityJpaController friends;
    private UserInfoElementEntityJpaController userInfos;

    public JPADatabase() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory("GoodGameServerPU");
        users = new UserEntityJpaController(emf);
        gameTypes = new GameTypeEntityJpaController(emf);
        userResults = new UserResultsEntityJpaController(emf);
        friends = new FriendEntityJpaController(emf);
        userInfos = new UserInfoElementEntityJpaController(emf);

        List<GameTypeEntity> gameTypeList = gameTypes.findGameTypeEntityEntities();
        if (gameTypeList.isEmpty()) {
            GameTypeEntity desdemona = new GameTypeEntity();
            desdemona.setName("Desdemona");
            desdemona.setServerJarNames("Desdemona.jar,Dummy.jar");
            desdemona.setClientJarNames("Desdemona.jar,Dummy.jar");
            desdemona.setServerClassName("net.sf.goodgame.tbgf.games.desdemona.server.DesdemonaGameType");
            desdemona.setClientClassName("net.sf.goodgame.tbgf.games.desdemona.client.DesdemonaClient");
            GameTypeEntity domination = new GameTypeEntity();
            domination.setName("Domination");
            domination.setServerJarNames("Domination.jar");
            domination.setServerClassName("net.sf.goodgame.tbgf.emulator.server.DominationGameType");
            domination.setClientJarNames("Domination.jar");
            domination.setClientClassName("net.sf.goodgame.tbgf.emulator.client.LobbyGameEmulator");
            try {
                gameTypes.create(desdemona);
                gameTypes.create(domination);
                gameTypeList = new LinkedList();
                gameTypeList.add(desdemona);
                gameTypeList.add(domination);
            } catch (PreexistingEntityException ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            } catch (Exception ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        for (GameTypeEntity gameType : gameTypeList) {
            try {
                //GameType gt = new GameType(gameType.getName(), gameType.getServerJarNames(), gameType.getServerClassName(), gameType.getClientJarNames(), gameType.getClientClassName());
                GameType gt = GameType.getGameType(gameType.getServerJarNames(), gameType.getServerClassName(), gameType.getClientJarNames(), gameType.getClientClassName());
            } catch (Exception e) {
                System.out.println("Error creating game type: " + gameType.getName());
                e.printStackTrace();
            }
        }
        List<UserEntity> allUsers = users.findUserEntityEntities();
        for (UserEntity user : allUsers) {
            if (!user.isLoggedIn()) {
                continue;
            }
            user.setLoggedIn(false);
            try {
                users.edit(user);
            } catch (NonexistentEntityException ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            } catch (Exception ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    @Override
    public synchronized void addFriend(User user, String friend) {
        FriendEntity myFriend = new FriendEntity();
        myFriend.setUsername(user.getUsername());
        myFriend.setFriend(friend);
        try {
            friends.create(myFriend);
        } catch (PreexistingEntityException ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        } catch (Exception ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public synchronized void removeFriend(User user, String friend) {
        FriendEntityID id = new FriendEntityID();
        id.setFriend(friend);
        id.setUsername(user.getUsername());
        try {
            friends.destroy(id);
        } catch (NonexistentEntityException ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public synchronized FriendsList getFriendsList(User user) {
        List<FriendEntity> friendsEntities = friends.findFriends(user.getUsername());
        Collection<String> myFriends = new LinkedList<String>();
        if (friendsEntities == null) {
            System.out.println("Did not find any friends");
            return new FriendsList(user, myFriends);
        }

        for (FriendEntity friend : friendsEntities) {
            myFriends.add(friend.getFriend());
        }
        return new FriendsList(user, myFriends);

    }

    @Override
    public synchronized Collection<String> getFriendsOfList(String username) {
        List<FriendEntity> friendsEntities = friends.findFriendsOf(username);
        Collection<String> friendsOf = new LinkedList<String>();
        if (friendsEntities == null) {
            System.out.println(username + " is not a friend of anyone");
            return friendsOf;
        }
        for (FriendEntity friendOf : friendsEntities) {
            friendsOf.add(friendOf.getUsername());
        }
        return friendsOf;
    }

    @Override
    public synchronized UserType getType(String username) {
        UserEntity userEntity = users.findUserEntity(username);
        return userEntity.getUserType();
    }

    @Override
    public void saveUserInfo(String username, UserInfoElement info) {
        UserInfoElementEntityID id = new UserInfoElementEntityID();
        id.setUsername(username);
        id.setColumnName(info.getColumnName());
        UserInfoElementEntity entity = userInfos.findUserInfoElementEntity(id);
        if (entity == null) {
            //Create it.
            entity = new UserInfoElementEntity();
            entity.setUsername(username);
            entity.setColumnName(info.getColumnName());
            entity.setInfoText(info.getText());
            try {
                userInfos.create(entity);
            } catch (PreexistingEntityException ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            } catch (Exception ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        } else {
            entity.setInfoText(info.getText());
            try {
                userInfos.edit(entity);
            } catch (NonexistentEntityException ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            } catch (Exception ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    @Override
    public Collection<UserInfoElement> getUserInfo(String username) {
        List<UserInfoElementEntity> userInfoEntity = userInfos.findUserInfo(username);
        if (userInfoEntity == null) {
            userInfoEntity = new ArrayList<UserInfoElementEntity>();
        }
        List<UserResultsEntity> userResultsEntity = userResults.findUserResultsEntity(username);
        if (userResultsEntity == null) {
            userResultsEntity = new ArrayList<UserResultsEntity>();
        }
        List<UserInfoElement> userInfo = new ArrayList<UserInfoElement>();
        for (UserInfoElementEntity entity : userInfoEntity) {
            userInfo.add(new UserInfoElement(entity.getColumnName(),
                    entity.getInfoText(), UserInfoType.userSupplied));
        }
        for (UserResultsEntity results : userResultsEntity) {
            userInfo.add(new UserInfoElement(results.getGameType(), results.toString(),
                    UserInfoType.gameResults));
        }
        for (UserInfoElement element : userInfo) {
            System.out.println("ELEMENT: " + element);
        }
        return userInfo;
    }

    @Override
    public void logGameMove(GameType gameID, String user, String move) {
    }

    @Override
    public synchronized boolean login(String username, String password) {
        UserEntity userEntity = users.findUserEntity(username.trim().toLowerCase());
        if (userEntity == null) {
            System.out.println("Not registered");
            return false;
        }
        byte[] hash1 = Util.sha(password);
        byte[] hash2 = userEntity.getPassword();
        if (!MessageDigest.isEqual(hash1, hash2)) {
            System.out.println("\nInvalid password");
            return false;
        }
        if (userEntity.isLoggedIn()) {
            System.out.println("Already Logged in");
            return false;
        }
        userEntity.setLastLogin(new Timestamp(System.currentTimeMillis()));
        userEntity.setLoggedIn(true);
        try {
            users.edit(userEntity);
            return true;
        } catch (NonexistentEntityException ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        } catch (Exception ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
        System.out.println("Exception while logging in");
        return false;
    }

    @Override
    public synchronized void logout(String username) {
        UserEntity userEntity = users.findUserEntity(username);
        if (userEntity == null) {
            return;
        }
        userEntity.setLoggedIn(false);
        try {
            users.edit(userEntity);
        } catch (NonexistentEntityException ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        } catch (Exception ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public synchronized boolean registerUser(String username, String password, String email, UserType type) {

        if (userIsRegistered(username)) {
            return false;
        }
        UserEntity userEntity = new UserEntity();
        userEntity.setUsername(username);
        if (password.length() < 6) {
            System.out.println("Password is too short: "+password.length());
            return false;
        }
        byte[] passwordHash = Util.sha(password);
        userEntity.setPassword(passwordHash);
        userEntity.setEmailAddress(email);
        userEntity.setUserType(type);
        userEntity.setLoggedIn(false);
        userEntity.setLastLogin(new Timestamp(System.currentTimeMillis()));
        try {
            users.create(userEntity);
            return true;
        } catch (PreexistingEntityException ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        } catch (Exception ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        }

        System.out.println("Failed registration: registered? " + userEntity);
        return false;
    }

    @Override
    public synchronized void storeResults(GameType gt, GameResult r) {
        if (r.countsTowardsStats()) {
            if (!r.wasTied()) {
                for (User winner : r.getWinners()) {
                    storeResult(winner, gt, false, true);
                }
                for (User loser : r.getLosers()) {
                    storeResult(loser, gt, false, false);
                }
            } else {
                for (User tie : r.getTie()) {
                    storeResult(tie, gt, true, false);
                }
            }
        }
    }

    public synchronized Results getResults(GameType gt, User u) {
        UserResultsEntityID id = new UserResultsEntityID();
        id.setGameType(gt.NAME);
        id.setUsername(u.getUsername());
        UserResultsEntity results = userResults.findUserResultsEntity(id);
        return results;
    }

    private void storeResult(User user, GameType gameType, boolean tie, boolean win) {
        UserResultsEntityID id = new UserResultsEntityID();
        id.setGameType(gameType.NAME);
        id.setUsername(user.getUsername());
        UserResultsEntity resultsEntity = userResults.findUserResultsEntity(id);
        if (resultsEntity == null) {
            resultsEntity = new UserResultsEntity();
            resultsEntity.setGameType(gameType.NAME);
            resultsEntity.setUsername(user.getUsername());
            modifyResult(resultsEntity, tie, win);
            try {
                userResults.create(resultsEntity);
            } catch (PreexistingEntityException ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            } catch (Exception ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        } else {
            modifyResult(resultsEntity, tie, win);
            try {
                userResults.edit(resultsEntity);
            } catch (NonexistentEntityException ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            } catch (Exception ex) {
                Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    private void modifyResult(UserResultsEntity results, boolean tie, boolean win) {
        if (tie) {
            results.setTies(results.getTies() + 1);
        } else if (win) {
            results.setWins(results.getWins() + 1);
        } else {
            results.setLosses(results.getLosses() + 1);
        }
    }

    @Override
    public synchronized boolean userIsRegistered(String username) {
        UserEntity userEntity = users.findUserEntity(username.trim().toLowerCase());
        return userEntity != null;
    }

    public void storeGameType(GameType gt) {
        GameTypeEntity gameTypeEntity = new GameTypeEntity();
        gameTypeEntity.setName(gt.NAME);
        gameTypeEntity.setClientClassName(gt.getClientClassName());
        gameTypeEntity.setClientJarNames(gt.getClientJarNames());
        gameTypeEntity.setServerClassName(gt.getServerClassName());
        gameTypeEntity.setServerJarNames(gt.getServerJarNames());
        try {
            gameTypes.create(gameTypeEntity);
        } catch (PreexistingEntityException ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        } catch (Exception ex) {
            Logger.getLogger(JPADatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public String[] getRegisteredUsers() {
        List<UserEntity> entitys = users.findUserEntityEntities();
        String[] registered = new String[entitys.size()];
        int index = 0;
        Iterator<UserEntity> iter = entitys.iterator();
        while(iter.hasNext())
        {
            registered[index++] = iter.next().getUsername();
        }
        return registered;
    }

    public void close() {
        friends.getEntityManager().close();
        gameTypes.getEntityManager().close();
        userInfos.getEntityManager().close();
        users.getEntityManager().close();
        userResults.getEntityManager().close();
    }

}
