/*
(C) 2008 Michael Chaten
This file is part of GoodGame.

GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License

GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.server.persistence.jpa;

import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Query;
import javax.persistence.EntityNotFoundException;
import net.sf.goodgame.server.persistence.jpa.exceptions.NonexistentEntityException;
import net.sf.goodgame.server.persistence.jpa.exceptions.PreexistingEntityException;
import org.apache.openjpa.persistence.InvalidStateException;

/**
 *
 * @author mchaten
 */
public class FriendEntityJpaController {

    private EntityManagerFactory emf = null;

    public FriendEntityJpaController(EntityManagerFactory emf) {
        this.emf = emf;
    }

    public EntityManager getEntityManager() {
        return emf.createEntityManager();
    }

    public void create(FriendEntity friendEntity) throws PreexistingEntityException, Exception {
        EntityManager em = null;
        try {
            em = getEntityManager();
            em.getTransaction().begin();
            em.persist(friendEntity);
            em.getTransaction().commit();
        } catch (Exception ex) {
            FriendEntityID id = new FriendEntityID(friendEntity);
            if (findFriendEntity(id) != null) {
                throw new PreexistingEntityException("FriendEntity " + friendEntity + " already exists.", ex);
            }
            throw ex;
        } finally {
            if (em != null) {
                try {
                    em.close();
                } catch (InvalidStateException e) {
                    System.out.println("Failed object: " + e.getFailedObject());
                    throw e;
                }
            }
        }
    }

    public void edit(FriendEntity friendEntity) throws NonexistentEntityException, Exception {
        EntityManager em = null;
        try {
            em = getEntityManager();
            em.getTransaction().begin();
            friendEntity = em.merge(friendEntity);
            em.getTransaction().commit();
        } catch (Exception ex) {
            String msg = ex.getLocalizedMessage();
            if (msg == null || msg.length() == 0) {
                FriendEntityID id = new FriendEntityID(friendEntity);
                if (findFriendEntity(id) == null) {
                    throw new NonexistentEntityException("The friendEntity with id " + id + " no longer exists.");
                }
            }
            throw ex;
        } finally {
            if (em != null) {
                em.close();
            }
        }
    }

    public void destroy(FriendEntityID id) throws NonexistentEntityException {
        EntityManager em = null;
        try {
            em = getEntityManager();
            em.getTransaction().begin();
            FriendEntity friendEntity;
            try {
                friendEntity = em.getReference(FriendEntity.class, id);
                friendEntity.getUsername();
                friendEntity.getFriend();
            } catch (EntityNotFoundException enfe) {
                throw new NonexistentEntityException("The friendEntity with id " + id + " no longer exists.", enfe);
            }
            em.remove(friendEntity);
            em.getTransaction().commit();
        } finally {
            if (em != null) {
                em.close();
            }
        }
    }

    public List<FriendEntity> findFriendEntityEntities() {
        return findFriendEntityEntities(true, -1, -1);
    }

    public List<FriendEntity> findFriendEntityEntities(int maxResults, int firstResult) {
        return findFriendEntityEntities(false, maxResults, firstResult);
    }

    private List<FriendEntity> findFriendEntityEntities(boolean all, int maxResults, int firstResult) {
        EntityManager em = getEntityManager();
        try {
            Query q = em.createQuery("select object(o) from FriendEntity as o");
            if (!all) {
                q.setMaxResults(maxResults);
                q.setFirstResult(firstResult);
            }
            return q.getResultList();
        } finally {
            em.close();
        }
    }

    public List<FriendEntity> findFriends(String username) {
        EntityManager em = getEntityManager();
        try {
            Query q = em.createQuery("select object(o) from FriendEntity as o where o.username=?1").setParameter(1, username);
            return q.getResultList();
        } finally {
            em.close();
        }
    }

    public List<FriendEntity> findFriendsOf(String username) {
        EntityManager em = getEntityManager();
        try {
            Query q = em.createQuery("select object(o) from FriendEntity as o where o.friend=?1").setParameter(1, username);
            return q.getResultList();
        } finally {
            em.close();
        }
    }

    public FriendEntity findFriendEntity(FriendEntityID id) {
        EntityManager em = getEntityManager();
        try {
            return em.find(FriendEntity.class, id);
        } finally {
            em.close();
        }
    }

    public int getFriendEntityCount() {
        EntityManager em = getEntityManager();
        try {
            return ((Long) em.createQuery("select count(o) from FriendEntity as o").getSingleResult()).intValue();
        } finally {
            em.close();
        }
    }
}
