/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.sf.goodgame.shared.network;

import java.security.InvalidParameterException;
import net.sf.goodgame.shared.UserInfoElement;
import net.sf.goodgame.shared.games.GameState;
import net.sf.goodgame.shared.UserType;
import net.sf.goodgame.shared.Util;
import net.sf.goodgame.shared.games.GameOptions;
import net.sf.goodgame.shared.games.table.GameTableColumn;
import net.sf.goodgame.shared.games.table.GameTableLayout;
import net.sf.goodgame.shared.games.table.GameTableRow;

/**
 *
 * @author Mike
 */
public abstract class NetworkMessage {

    public abstract byte[] getPayload();

    public abstract byte[] getHeader();

    public abstract void decrypt();

    public abstract void encrypt();

    public abstract Object[] decodeMessage(boolean list, Class... data) throws InvalidMessageException;

    public static NetworkMessage createMessage(ServerToClientHeader header, Object... dataArray) {
        byte[] headerArray = new byte[ServerToClientHeader.HEADER_SIZE];
        assert header.messageNumberArray.length == 2;
        headerArray[0] = header.messageNumberArray[0];
        headerArray[1] = header.messageNumberArray[1];
        byte[] payload = getPayload(dataArray);
        if (payload.length > Short.MAX_VALUE) {
            return null;
        }
        byte[] payloadLen = Util.shortToByteArray((short) payload.length);
        assert headerArray.length == ServerToClientHeader.HEADER_SIZE ||
                headerArray.length == ClientToServerHeader.HEADER_SIZE;
        headerArray[2] = payloadLen[0];
        headerArray[3] = payloadLen[1];
        return new ImmutableNetworkMessage(header, headerArray, payload);
    }

    public static NetworkMessage createMessage(ClientToServerHeader header, Object... dataArray) {
        byte[] headerArray = new byte[ClientToServerHeader.HEADER_SIZE];
        assert header.messageNumberArray.length == 2;
        headerArray[0] = header.messageNumberArray[0];
        headerArray[1] = header.messageNumberArray[1];
        byte[] payload = getPayload(dataArray);
        if (payload.length > Short.MAX_VALUE) {
            return null;
        }
        byte[] messageLen = Util.shortToByteArray((short) payload.length);
        assert headerArray.length == ServerToClientHeader.HEADER_SIZE ||
                headerArray.length == ClientToServerHeader.HEADER_SIZE;
        headerArray[2] = messageLen[0];
        headerArray[3] = messageLen[1];
        return new ImmutableNetworkMessage(header, headerArray, payload);
    }

    private static byte[] getPayload(Object[] dataArray) {
        byte[][] dataBytes = new byte[dataArray.length][];
        for (int i = 0; i < dataArray.length; i++) {
            Object data = dataArray[i];
            if (data instanceof String) {
                dataBytes[i] = stringPayload((String) data);
            } else if (data instanceof Byte) {
                dataBytes[i] = bytePayload((Byte) data);
            } else if (data instanceof Short) {
                dataBytes[i] = shortPayload((Short) data);
            } else if (data instanceof Integer) {
                dataBytes[i] = intPayload((Integer) data);
            } else if (data instanceof GameOptions) {
                if (i + 1 != dataArray.length) {
                    throw new InvalidParameterException("GameOptions must be at the end of the message");
                }
                dataBytes[i] = ((GameOptions) data).getNetworkMessage();
            } else if (data instanceof GameState) {
                dataBytes[i] = new byte[]{((GameState) data).networkNumber};
            } else if (data instanceof byte[]) {
                dataBytes[i] = byteArrayPayload((byte[]) data);
            } else if (data instanceof UserType) {
                byte[] array = new byte[]{((UserType) data).privileges};
                dataBytes[i] = array;
            } else if (data instanceof GameTableRow) {
                dataBytes[i] = gameTableRowPayload((GameTableRow) data);
            } else if (data instanceof GameTableLayout) {
                dataBytes[i] = gameTableLayoutPayload((GameTableLayout) data);
            } else if(data instanceof UserInfoElement) {
                dataBytes[i] = userInfoElementPayload((UserInfoElement)data);
            } else if(data instanceof UserInfoElement[]) {
                dataBytes[i] = userInfoElementArrayPayload((UserInfoElement[])data);
            }
            else {
                throw new InvalidParameterException("Unknown data type: " + data.getClass());
            }
        }

        return Util.combineArrays(dataBytes);
    }

    private static byte[] stringPayload(String data) {
        String string = (String) data;
        byte[] bytes = string.getBytes();
        short strLen = (short) bytes.length;
        return Util.combineArrays(Util.shortToByteArray(strLen), bytes);
    }

    private static byte[] bytePayload(byte data) {
        return new byte[]{((Byte) data)};
    }

    private static byte[] shortPayload(short data) {
        return Util.shortToByteArray((Short) data);
    }

    private static byte[] intPayload(int data) {
        return Util.intToByteArray((Integer) data);
    }

    private static byte[] byteArrayPayload(byte[] data) {
        byte[] array = (byte[]) data;
        if (array.length > Short.MAX_VALUE) {
            throw new InvalidParameterException("Byte Array too large");
        }
        return Util.combineArrays(Util.shortToByteArray((short) array.length), array);
    }

    private static byte[] gameTableRowPayload(GameTableRow data) {
        short len = 0;
        String[] columns = ((GameTableRow) data).getColumns();
        byte[][] array = new byte[columns.length][];
        for (int j = 0; j < columns.length; j++) {
            byte[] column = columns[j].getBytes();
            if (len + column.length > Short.MAX_VALUE) {
                throw new InvalidParameterException("Column length too large");
            }
            len += column.length + 2;
            array[j] = Util.combineArrays(Util.shortToByteArray((short) column.length), column);
        }
        return Util.combineArrays(Util.shortToByteArray(len), Util.combineArrays(array));
    }

    private static byte[] gameTableLayoutPayload(GameTableLayout data) {
        short len = 0;
        GameTableColumn[] columns = ((GameTableLayout) data).getColumns();
        byte[][] array = new byte[columns.length][];
        for (int j = 0; j < columns.length; j++) {
            byte[] name = columns[j].getName().getBytes();
            if (len + name.length > Short.MAX_VALUE) {
                throw new InvalidParameterException("Name length too large");
            }
            short nameLen = (short) name.length;
            len += nameLen + 3;
            array[j] = Util.combineArrays(new byte[]{columns[j].getSize()}, Util.shortToByteArray(nameLen), name);
        }
        return Util.combineArrays(Util.shortToByteArray(len), Util.combineArrays(array));
    }
    private static byte[] userInfoElementPayload(UserInfoElement data) {
        byte isEditable = data.getType().getOrdinal();
        byte[] columnName = stringPayload(data.getColumnName());
        byte[] text = stringPayload(data.getText());
        return Util.combineArrays(columnName,text,new byte[]{isEditable});
    }
    private static byte[] userInfoElementArrayPayload(UserInfoElement[] data) {
        byte[][] array = new byte[data.length+1][];
        array[0] = Util.shortToByteArray((short)data.length);//Num Elements in the array
        for(int i = 0;i<data.length;i++)
        {
            array[i+1] = userInfoElementPayload(data[i]);
        }
        return Util.combineArrays(array);
    }

    @Override
    public abstract String toString();
}
